import 'dart:async';
import 'dart:convert';

import 'package:convert/convert.dart';
import 'package:crypto/crypto.dart';
import 'package:fake_http/fake_http.dart';
import 'package:flutter/foundation.dart';

/// 网络层拦截器
abstract class SignInterceptor implements Interceptor {
  static const String KEY_SIGN = "sign"; // 签名MD5 32 位 大写
  //static const String KEY_CERT = "cert";// ipa签名校验
  static const String KEY_UUID = "uuid"; // 设备唯一标识
  static const String KEY_TIME = "t"; // 时间戳
  static const String KEY_APPID = "appid";

  Future<Response> interceptSign(
    Chain chain,
    String appId,
    String appSecret,
    AsyncValueGetter<String> uuid,
  ) async {
    Request originalRequest = chain.request();
    Headers originalHeaders = originalRequest.headers();
    HttpUrl originalHttpUrl = originalRequest.url();

    HeadersBuilder signHeadersBuilder = originalHeaders.newBuilder();
    signHeadersBuilder.set(KEY_APPID, appId);

    HttpUrlBuilder signHttpUrlBuilder = originalHttpUrl.newBuilder();
    signHttpUrlBuilder
        .removeAllQueryParameters(KEY_UUID)
        .addQueryParameter(KEY_UUID, await uuid());
    signHttpUrlBuilder.removeAllQueryParameters(KEY_TIME).addQueryParameter(
        KEY_TIME, DateTime.now().millisecondsSinceEpoch.toString());
    signHttpUrlBuilder
        .removeAllQueryParameters(KEY_APPID)
        .addQueryParameter(KEY_APPID, appId);
    signHttpUrlBuilder.removeAllQueryParameters(KEY_SIGN).addQueryParameter(
        KEY_SIGN,
        Sign.IMPL.generate(appId, appSecret, signHttpUrlBuilder.build()));

    Response signResponse = await chain.proceed(originalRequest
        .newBuilder()
        .headers(signHeadersBuilder.build())
        .url(signHttpUrlBuilder.build())
        .build());

    return signResponse.newBuilder().request(originalRequest).build();
  }

  Response interceptError(Chain chain, String message) {
    return ResponseBuilder()
        .request(chain.request())
        .code(HttpStatus.gatewayTimeout)
        .message(message)
        .body(Util.EMPTY_RESPONSE)
        .sentRequestAtMillis(-1)
        .receivedResponseAtMillis(DateTime.now().millisecondsSinceEpoch)
        .build();
  }
}

abstract class Sign {
  static const Sign IMPL = _SortSign();

  String generate(String appId, String appSecret, HttpUrl httpUrl);
}

class _SortSign implements Sign {
  const _SortSign();

  @override
  String generate(String appId, String appSecret, HttpUrl httpUrl) {
    Map<String, String> metaMap = <String, String>{};
    List<String> parameterNames = httpUrl.queryParameterNames().toList();
    parameterNames.sort(); // sort by key [0-9][A-Z][a-z]
    for (String parameterName in parameterNames) {
      metaMap.putIfAbsent(
          parameterName, () => httpUrl.queryParameter(parameterName));
    }
    StringBuffer builder = StringBuffer();
    builder.write(appSecret);
    metaMap.forEach((String key, String value) {
      builder.write(key);
      builder.write(value);
    });
    return hex
        .encode(md5.convert(utf8.encode(builder.toString())).bytes)
        .toUpperCase();
  }
}

class SimpleSignInterceptor extends SignInterceptor {
  SimpleSignInterceptor(
    String appId,
    String appSecret,
    AsyncValueGetter<String> uuid,
  )   : assert(appId != null),
        assert(appSecret != null),
        assert(uuid != null),
        _appId = appId,
        _appSecret = appSecret,
        _uuid = uuid;

  final String _appId;
  final String _appSecret;
  final AsyncValueGetter<String> _uuid;

  @override
  Future<Response> intercept(Chain chain) {
    return interceptSign(chain, _appId, _appSecret, _uuid);
  }
}
